/*
Copyright (C) 2013  Xyne

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
(version 2) as published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#define STDIN "/dev/stdin"
#define STDOUT "/dev/stdout"
#define RANDOM "/dev/urandom"

/*
  TODO
  Consider moving most of this to a man page.
*/
void print_usage(char * name)
{
  fprintf(
    stderr,
    "ABOUT\n"
    "  %s - split and join files via cryptologically secure one-time pads\n"
    "\n"
    "  There are two splitting modes: random and pre-existing pad.\n"
    "\n"
    "  The random mode uses a random input source to generate (n-1) of the given n\n"
    "  output files. The final output file is computed from the bitwise compliment of\n"
    "  the other generated files. Taken together, the (n-1) generated files may be\n"
    "  considered as a one-time-pad for the encrypted data, although the distinction\n"
    "  of which file is the encrypted data is trivial.\n"
    "\n"
    "  The pre-existing pad mode uses an existing file to encrypt the data. This mode\n"
    "  will only generate one output file, which is the file encrypted with the pad\n"
    "  data.\n"
    "\n"
    "  There is only one joining mode. Randomly split files may be recovered by\n"
    "  passing all of them in any order. A file encrypted with a pre-existing pad may\n"
    "  be recovered by passing it first followed by the one-time-pad that was used to\n"
    "  create it. The order is important if the one-time-pad is not the exact same\n"
    "  byte length as the encrypted file.\n"
    "\n"
    "USAGE\n"
    "  %s -s [-r <path to random input>] <filepath> <filepath> [<filepath>...]\n"
    "  %s -j <filepath> <filepath> [<filepath>...]\n"
    "\n"
    "  %s -s -o <path to one-time-pad> <path to plaintext file>\n"
    "  %s -j  <path to encrypted file> <path to one-time-pad>\n"
    "\n"
    "  -h            : display this message and exit\n"
    "  -j            : join mode\n"
    "  -o <filepath> : use existing one-time pad\n"
    "  -r <filepath> : input device for random data (default: %s)\n"
    "  -s            : split mode (default)\n"
    "\n"
    "  \"-\" may be passed as a special filepath argument. When splitting it will\n"
    "  redirect output to STDOUT. When joining it will read input from STDIN.\n"
    "\n"
    "EXAMPLES\n"
    "  %s -s foo.a foo.b foo.c foo.d < bar\n"
    "  %s -j foo.a foo.b foo.c foo.d > bar.copy\n"
    "\n"
    "  %s -s -o pad foo < bar\n"
    "  %s -j foo pad > bar.copy\n"
    "\n"
    "DISCLAIMER\n"
    "  This program is distributed in the hope that it will be useful,\n"
    "  but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
    "  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE,\n"
    "\n"
    "  If you require reliable cryptography, check the source code yourself to\n"
    "  ensure that it satisfies your requirements.\n"
    "\n"
    "AUTHOR\n"
    "  Xyne, 2013\n"
    "\n",
    name,
    name,
    name,
    name,
    name,
    RANDOM,
    name,
    name,
    name,
    name
  );
}

int main(int argc, char * * argv)
{
  int i = 0, n, split = 1, last_count;
  char * random = RANDOM, * one_time_pad = NULL;
  long long unsigned int data = 0, noise = 0, sum = 0;
  size_t count = 0;
  FILE * random_fh;

  /*
    Basic option handling to select split and join modes.
  */
  while ((i = getopt(argc, argv, "hjo:r:s")) != -1)
  {
    switch (i)
    {
      case 'h':
        print_usage(argv[0]);
        return EXIT_SUCCESS;
      case 'j':
        split = 0;
        break;
      case 'o':
        one_time_pad = optarg;
        break;
      case 'r':
        random = optarg;
        break;
      case 's':
        split = 1;
        break;
      case '?':
        return EXIT_FAILURE;
    }
  }

  if (
    optind == argc || \
    (
      one_time_pad != NULL && \
      (argc - optind) > 1)
    )
  {
    print_usage(argv[0]);
    return EXIT_FAILURE;
  }

  n = argc - optind;
  stdin = freopen(STDIN, "rb", stdin);

  FILE * fhs[n];

  for (i=optind; i<argc; i++)
  {
    if (argv[i][0] == '-' && argv[i][1] == '\0')
    {
      fhs[i-optind] = split ? stdout : stdin;
    }
    else if ((fhs[i-optind] = fopen(argv[i], split ? "wb" : "rb")) == NULL)
    {
      perror(argv[0]);
      return EXIT_FAILURE;
    }
  }

  if (split)
  {
    if (one_time_pad != NULL)
    {
      random_fh = fopen(one_time_pad, "rb");
    }
    else
    {
      random_fh = fopen(random, "rb");
    }
    if (random_fh == NULL)
    {
      perror(argv[0]);
      return EXIT_FAILURE;
    }

    if (stdin == NULL)
    {
      fprintf(stderr, "error: failed to obtain input from STDIN\n");
      return EXIT_FAILURE;
    }

    data = 0;
    while ((count = fread(&data, 1, sizeof(data), stdin)) > 0)
    {
      sum = 0;
      if (one_time_pad == NULL)
      {
        for (i=0; i<(n-1); i++)
        {
          if (fread(&noise, 1, sizeof(noise), random_fh) != sizeof(data))
          {
            perror(argv[0]);
            return EXIT_FAILURE;
          }
          if (fwrite(&noise, 1, count, fhs[i]) != count)
          {
            perror(argv[0]);
            return EXIT_FAILURE;
          }
          sum += noise;
        }
        noise = data - sum;
      }
      else
      {
        i = 0;
        if (fread(&noise, 1, count, random_fh) != count)
        {
          if (errno)
          {
            perror(argv[0]);
          }
          else
          {
            fprintf(stderr, "error: pad is too short\n");
          }
          return EXIT_FAILURE;
        }
        noise = data - noise;
      }
      if (fwrite(&noise, 1, count, fhs[i]) != count)
      {
        perror(argv[0]);
        return EXIT_FAILURE;
      }
      data = 0;
    }
    fclose(random_fh);
    fclose(stdin);
  }

  // join
  else
  {
//     stdout = freopen(STDOUT, "wb", stdout);
    while (1)
    {
      last_count = -1;
      count = sizeof(noise);
      sum = 0;
      for (i=0; i<n; i++)
      {
        count = fread(&noise, 1, count, fhs[i]);
        if (last_count < 0)
        {
          last_count = count;
        }
        else if (count != last_count)
        {
          if (errno)
          {
            perror(argv[0]);
          }
          else
          {
            fprintf(stderr, "error: file %d is too short\n", i);
          }
          return EXIT_FAILURE;
        }
        sum += noise;
      }
      if (count == 0)
      {
        break;
      }
      if (fwrite(&sum, 1, count, stdout) != count)
      {
        perror(argv[0]);
        return EXIT_FAILURE;
      }
    }
    fclose(stdout);
  }

  for (i=0; i<n; i++)
  {
    fclose(fhs[i]);
  }

  return EXIT_SUCCESS;
}
